--- win_efile.c	2009-09-02 17:32:03.140734600 -0700
+++ ../otp/win_efile.c	2009-09-02 17:48:48.423423500 -0700
@@ -778,6 +778,8 @@
 efile_fileinfo(Efile_error* errInfo, Efile_info* pInfo,
 	       char* orig_name, int info_for_link)
 {
+	HANDLE handle;			/* Handle returned by CreateFile()	   */
+	BY_HANDLE_FILE_INFORMATION fileInfo; /* from  CreateFile()	   */
     HANDLE findhandle;		/* Handle returned by FindFirstFile(). */
     WIN32_FIND_DATA findbuf;	/* Data return by FindFirstFile(). */
     char name[_MAX_PATH];
@@ -849,6 +851,7 @@
         findbuf.nFileSizeLow = 0;
         findbuf.cFileName[0] = '\0';
 
+	pInfo->links = 1;
 	pInfo->modifyTime.year = 1980;
 	pInfo->modifyTime.month = 1;
 	pInfo->modifyTime.day = 1;
@@ -861,6 +864,28 @@
 	SYSTEMTIME SystemTime;
         FILETIME LocalFTime;
 
+		//first check if we are a symlink:
+		if (!info_for_link && (findbuf.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT))
+		{
+			//given that we know this is a symlink, we should be able to find its target:
+			char target_name[256];
+			if (efile_readlink(errInfo, name, target_name,256) == 1)
+				return efile_fileinfo(errInfo, pInfo,target_name, info_for_link);
+			//else, continue on and read this file:
+		}
+
+		//number of links:
+		if(handle = CreateFile(name, GENERIC_READ, 0,NULL, OPEN_EXISTING, 0, NULL))
+		{
+			GetFileInformationByHandle(handle, &fileInfo);
+			pInfo->links = fileInfo.nNumberOfLinks;
+
+			CloseHandle(handle);
+		}
+		else
+			pInfo->links = 1;
+		//---------------------------------------
+		
 #define GET_TIME(dst, src) \
 if (!FileTimeToLocalFileTime(&findbuf.src, &LocalFTime) || \
     !FileTimeToSystemTime(&LocalFTime, &SystemTime)) { \
@@ -895,7 +920,9 @@
     pInfo->size_low = findbuf.nFileSizeLow;
     pInfo->size_high = findbuf.nFileSizeHigh;
 	
-    if (findbuf.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
+	if (info_for_link && (findbuf.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT))
+	pInfo->type = FT_SYMLINK;
+    else if (findbuf.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
 	pInfo->type = FT_DIRECTORY;
     else
 	pInfo->type = FT_REGULAR;
@@ -906,7 +933,6 @@
 	pInfo->access = FA_READ|FA_WRITE;
 
     pInfo->mode = dos_to_posix_mode(findbuf.dwFileAttributes, name);
-    pInfo->links = 1;
     pInfo->major_device = drive;
     pInfo->minor_device = 0;
     pInfo->inode = 0;
@@ -1330,6 +1356,54 @@
 int
 efile_readlink(Efile_error* errInfo, char* name, char* buffer, size_t size)
 {
+	//load dll and see if we have CreateSymbolicLink at runtime:
+	//Vista only
+	HINSTANCE hModule = NULL;
+	if ((hModule = LoadLibrary("kernel32.dll")) != NULL)
+	{
+		typedef DWORD (WINAPI * GETFINALPATHNAMEBYHANDLEPTR)(
+			HANDLE hFile,
+			LPCSTR lpFilePath,
+			DWORD cchFilePath,
+			DWORD dwFlags
+		);
+
+		GETFINALPATHNAMEBYHANDLEPTR pGetFinalPathNameByHandle =
+			(GETFINALPATHNAMEBYHANDLEPTR)
+			GetProcAddress(hModule, "GetFinalPathNameByHandleA");
+
+		if (pGetFinalPathNameByHandle == NULL)			
+			FreeLibrary(hModule);
+		else 
+		{
+			//first check if file is a symlink ; {error, einval} otherwise
+			DWORD fileAttributes =  GetFileAttributes(name);
+			if ((fileAttributes & FILE_ATTRIBUTE_REPARSE_POINT))
+			{
+				BOOLEAN success = 0;
+				HANDLE h = CreateFile(name, GENERIC_READ, 0,NULL, OPEN_EXISTING, 0, NULL);
+				if(h != INVALID_HANDLE_VALUE)
+				{
+					success = pGetFinalPathNameByHandle(h, buffer, size,0);
+					//GetFinalPathNameByHandle prepends path with "\\?\":
+					sprintf(buffer, buffer+4);
+					CloseHandle(h);
+				}
+				FreeLibrary(hModule);
+
+				if(success)
+					return 1;
+				else
+					return set_error(errInfo);
+			}
+			else
+			{
+				FreeLibrary(hModule);
+				errno = EINVAL;
+				return check_error(-1, errInfo);
+			}
+		}
+	}
     errno = ENOTSUP;
     return check_error(-1, errInfo);
 }
@@ -1414,13 +1488,48 @@
 int
 efile_link(Efile_error* errInfo, char* old, char* new)
 {
-    errno = ENOTSUP;
-    return check_error(-1, errInfo);
+	if(!CreateHardLink(new, old, NULL))
+		return set_error(errInfo);
+
+	return 1;
 }
 
 int
 efile_symlink(Efile_error* errInfo, char* old, char* new)
 {
+	//load dll and see if we have CreateSymbolicLink at runtime:
+	//Vista only
+	HINSTANCE hModule = NULL;
+	if ((hModule = LoadLibrary("kernel32.dll")) != NULL)
+	{
+		typedef BOOLEAN (WINAPI  * CREATESYMBOLICLINKFUNCPTR)(
+			LPCSTR lpSymlinkFileName,
+			LPCSTR lpTargetFileName,
+			DWORD dwFlags
+		);
+
+		CREATESYMBOLICLINKFUNCPTR pCreateSymbolicLink =
+			(CREATESYMBOLICLINKFUNCPTR)
+			GetProcAddress(hModule, "CreateSymbolicLinkA");
+		//A for MBCS, W for UNICODE... char* above implies 'A'!
+
+		if (pCreateSymbolicLink != NULL)
+		{
+			DWORD attr = GetFileAttributes(old);
+			int flag = (attr != INVALID_FILE_ATTRIBUTES &&
+						attr & FILE_ATTRIBUTE_DIRECTORY) ?
+						  1 : 0; // SYMBOLIC_LINK_FLAG_DIRECTORY = 1
+			BOOLEAN success = pCreateSymbolicLink(new, old, flag);
+			FreeLibrary(hModule);
+
+			if(success)
+				return 1;
+			else
+				return set_error(errInfo);
+		}
+		else
+			FreeLibrary(hModule);
+	}
     errno = ENOTSUP;
     return check_error(-1, errInfo);
 }
